unit hiRarInfo;

interface

uses kol,windows,ShellAPI,Share,Debug;

const
  LocalFileHeaderSig = $04034b50;
  CentralFileHeaderSig = $02014b50;
  EndCentralDirSig = $06054b50;
  faDirectory = $00000010;
  ThousandSeparator = ' ';
  soFromCurrent = 1;
  soFromBeginning = 0;
  HoursPerDay = 24;
  MinsPerHour = 60;
  SecsPerMin  = 60;
  MSecsPerSec = 1000;
  MinsPerDay = HoursPerDay * MinsPerHour;
  SecsPerDay = MinsPerDay * SecsPerMin;
  MSecsPerDay = SecsPerDay * MSecsPerSec;

type
  LongRec = packed record
    case Integer of
      0: (Lo, Hi: Word);
      1: (Words: array [0..1] of Word);
      2: (Bytes: array [0..3] of Byte);
  end;

type
  TArcFile = record
    FileName    : STRING;
    DateTime    : TDateTime;
    Comp_Size   : DWord;
    UnComp_Size : DWord;
    Ratio       : Byte;
    ISProtected : Boolean;
    FileComment : STRING;
    OS          : Byte;
    OSStr       : STRING;
    CRC         : DWord;
    Method      : Byte;
    MethodStr   : STRING;
    Attr        : DWord;
    Unp_Ver     : Byte;
    Unp_VerStr  : STRING;
    FileType    : STRING;
    ImageIndex  : Integer;
  END;

type
  THIRarInfo = class(TDebug)
    FP: file of byte;
    Files: array[1..10000] of TArcFile;
    VolumeNum: Integer;
    NrFiles: Word;
    Offset: DWord;

    constructor Create;
    destructor Destroy;override;

    procedure Read(Value: String);
    function  GetByte: Byte;
    function  GetWord: Word;
    function  GetDWord: DWord;
    procedure ResetF(x: Word);
    function  CalcRatio(PackSize, UnCompSize: DWORD): Byte;

   public
    _event_onNext:THI_Event;
    _event_onEnd:THI_Event;

    procedure _work_doReadFileRar(var _Data:TData; Index:word);
    procedure _var_FileName(var _Data:TData; Index:word);
    procedure _var_DateTime(var _Data:TData; Index:word);
    procedure _var_Packed(var _Data:TData; Index:word);
    procedure _var_UnPacked(var _Data:TData; Index:word);
    procedure _var_CompressionMethod(var _Data:TData; Index:word);
    procedure _var_CRC32(var _Data:TData; Index:word);
    procedure _var_FileTypeInfo(var _Data:TData; Index:word);
  end;

implementation

constructor THIRarInfo.Create;
begin
  inherited Create;
end;

Destructor THIRarInfo.Destroy;
begin
  inherited Destroy;
end;

function GetFileTypeInfo(FileName: string): string;
var
  ShFileInfo: TShFileInfo;
  TypeInfo: string[79];
begin
  ShGetFileInfo(PChar(FileName), 0, ShFileInfo , SizeOf(TShFileInfo),
  SHGFI_SYSICONINDEX or  SHGFI_TYPENAME or SHGFI_USEFILEATTRIBUTES);
  TypeInfo := ShFileInfo.szTypeName;
  Result := TypeInfo;
end;

function Frmt(ts: string): string;
var
  a: Integer;
begin
  a := 3;
  while (a < length(ts)) do begin
    Insert(ThousandSeparator, ts, Length(ts) - a + 1); a := a + 4;
  end;
  Result := ts;
end;

function EncodeTime(Hour, Min, Sec, MSec: Word): TDateTime;
var Time: TDateTime;
begin
  if (Hour < HoursPerDay) and (Min < MinsPerHour) and (Sec < SecsPerMin) and (MSec < MSecsPerSec) then
  begin
    Time := (Hour * (MinsPerHour * SecsPerMin * MSecsPerSec) +
             Min * (SecsPerMin * MSecsPerSec) +
             Sec * MSecsPerSec +
             MSec) / MSecsPerDay;
  end;
  Result := Time;
end;

function FileDateToDateTime(FileDate: Integer): TDateTime;
var Date: TDateTime;
begin
    EncodeDate(
      LongRec(FileDate).Hi shr 9 + 1980,
      LongRec(FileDate).Hi shr 5 and 15,
      LongRec(FileDate).Hi and 31, Date);
    Result := Date + EncodeTime(
      LongRec(FileDate).Lo shr 11,
      LongRec(FileDate).Lo shr 5 and 63,
      LongRec(FileDate).Lo and 31 shl 1, 0);
end;

// Gets 1 byte (Char, Byte)
function THIRarInfo.GetByte : Byte;
var
  NumRead: Integer;
  B: Byte;
begin
  BlockRead(FP, B, 1, NumRead);
  Result := B;
end;

// Gets 2 bytes (Integer, Word)
function THIRarInfo.GetWord : Word;
var
  NumRead: Integer;
  W: Word;
begin
  BlockRead(FP, W, 2, NumRead);
  Result := W;
end;

// Gets 4 bytes (DWord)
function THIRarInfo.GetDWord : DWord;
var
  NumRead: Integer;
  W: DWord;
begin
  BlockRead(FP, W, 4, NumRead);
  Result := W;
end;

procedure THIRarInfo.ResetF(x: Word);
begin
  Files[x].FileName    := '';
  Files[x].DateTime    := 0;
  Files[x].Method      := 0;
  Files[x].IsProtected := False;
  Files[x].OS          := 0;
  Files[x].Attr        := 0;
  Files[x].Unp_VerStr  := '';
  Files[x].Unp_Ver     := 0;
  Files[x].CRC         := 0;
  Files[x].FileComment := '';
  Files[x].Ratio       := 0;
  Files[x].Comp_Size   := 0;
  Files[x].UnComp_Size := 0;
  Files[x].OSStr       := '';
  Files[x].MethodStr   := '';
  Files[x].FileType    := '';
  Files[x].ImageIndex  := 0;
end;

function THIRarInfo.CalcRatio(PackSize, UnCompSize: DWORD) : Byte;
begin
  if (UnCompSize > 0) AND (UnCompSize > PackSize) then Result := Round( (1 - (PackSize / UnCompSize)) * 100)
    else Result := 0;
end;

procedure THIRarInfo.Read(Value:String);
var
  i: Integer;
  b: Byte;
  w: Word;
  dw: DWord;
  hdr_size: Word;
  s: STRING[255];
  NextDisk: Boolean;
  CurrentName: STRING;
  PackSize: DWord;
  Cancel: Boolean;
label
  NewVolume;
begin
  NrFiles := 0;
  CurrentName := Value;

  AssignFile(FP, Value);
    Reset(FP);
    Seek(Fp, Offset);
    NewVolume:
    (* MARK_HEAD *)
    if (GetByte <> $52) then begin end;                (* HEAD_CRC *)
    if (GetByte <> $61) then begin end;                (* HEAD_CRC *)
    if (GetByte <> $72) then begin end;                (* HEAD_TYPE *)
    if (GetByte <> $21) then begin end;                (* HEAD_FLAGS *)
    if (GetByte <> $1a) then begin end;                (* HEAD_SIZE *)
    b := GetByte; b:= GetByte;

    (* MAIN_HEAD *)
    w := GetWord;                                      (* HEAD_CRC *)
    if (GetByte <> $73) then begin end;                (* HEAD_TYPE *)
    w := GetWord;                                      (* HEAD_FLAGS *)
    if (w AND $02 <> 0) then begin end;
    if (w AND $04 <> 0) then begin end;
    if (w AND $08 <> 0) then begin end;
    if (w AND $20 <> 0) then begin end;
    hdr_size := GetWord;                               (* HEAD_SIZE *)
    w := GetWord;                                      (* RESERVED1 *)
    dw := GetDWord;                                    (* RESERVED2 *)
    (* COMMENT BLOCK *)
    Seek(FP, FilePos(FP) + (hdr_size-13));

    (* FILE HEADER *)
    while (FilePos(FP) < FileSize(Value)) do begin
      NextDisk := False;
      w := GetWord;                                    (* HEAD_CRC *)
      if (GetByte <> $74 ) then begin exit end;        (* HEAD_TYPE *)
      Inc(NrFiles); ResetF(NrFiles);
      w := GetWord;                                    (* HEAD_FLAGS *)
      if (w AND $01 <> 0) then Dec(NrFiles);
      if (w AND $02 <> 0) then NextDisk := True;
      if (w AND $04 <> 0) then Files[NrFiles].IsProtected := True;
      if (w AND $08 <> 0) then begin end;
      if (w AND $10 <> 0) then begin end;
      if (w AND $01 <> 0) then begin end;

      hdr_size := GetWord;                             (* HEAD_SIZE *)

      dw := GetDWord;                                  (* PACK_SIZE *)
      PackSize := dw;
      Files[NrFiles].Comp_Size := Files[NrFiles].Comp_Size + dw;
      dw := GetDWord;                                  (* UNP_SIZE *)
      Files[NrFiles].UnComp_Size := dw;
      Files[NrFiles].Ratio := CalcRatio (Files[NrFiles].Comp_Size, Files[NrFiles].UnComp_Size);
      b  := GetByte;                                   (* HOST_OS *)
      Files[NrFiles].OS := b;
      case b of
        0 : Files[NrFiles].OSStr := 'DOS';
        1 : Files[NrFiles].OSStr := 'OS/2';
        2 : Files[NrFiles].OSStr := 'Win32';
        3 : Files[NrFiles].OSStr := 'Unix';
      end;

      dw := GetDWord;                                  (* FILE_CRC *)
      Files[NrFiles].CRC := dw;
      dw := GetDWord;                                  (* FTIME *)
      Files[NrFiles].DateTime := FileDateToDateTime(dw);
      b := GetByte;                                    (* UNP_VER *)
      Files[NrFiles].Unp_Ver := b;
      Files[NrFiles].Unp_VerStr := Int2Str(b DIV 10) + '.' + Int2Str(b MOD 10);
      b := GetByte;                                    (* METHOD              *)
      Files[NrFiles].Method := b;
      case b of
        $30 : Files[NrFiles].MethodStr := 'storing';
        $31 : Files[NrFiles].MethodStr := 'fastest';
        $32 : Files[NrFiles].MethodStr := 'fast';
        $33 : Files[NrFiles].MethodStr := 'normal';
        $34 : Files[NrFiles].MethodStr := 'good';
        $35 : Files[NrFiles].MethodStr := 'best';
        else
          Files[NrFiles].MethodStr := 'unknown';
      end;

      w := GetWord;                                    (* NAME_SIZE *)
      dw := GetDWord;                                  (* ATTR *)
      IF (dw AND faDirectory <> 0) then Files[NrFiles].FileName := Files[NrFiles].FileName + '\';
      Files[NrFiles].Attr := dw;

      for i:=1 to w do s[i] := Chr( GetByte );         (* FILE_NAME *)
      s[0] := Chr ( w );
      Files[NrFiles].FileName := s;

      (* COMMENT BLOCK *)
      Seek(FP, FilePos(FP) + (hdr_size-(32+w)));

      (* COMPRESSED FILE *)
      Seek(FP, FilePos(FP) + PackSize);

      (* PROCEED TO NEW VOLUME *)
      if (NextDisk) then begin
        Close(FP);
        if (VolumeNum = -1) then CurrentName[Length(Currentname)-2] := 'r';
          VolumeNum := VolumeNum + 1;
            if (VolumeNum = 100) then begin CurrentName[Length(CurrentName)-2] := Chr(Ord(CurrentName[Length(CurrentName)-2])+1); i:=0; END;
              Str(VolumeNum,s); if (VolumeNum<10) then s := '0'+s;
              CurrentName[Length(CurrentName)-1] := s[1];
              CurrentName[Length(CurrentName)] := s[2];
              Offset := 0;
              Cancel := True;
              if (NOT Cancel) then begin
                AssignFile(FP, CurrentName);
                Reset(FP);
                goto NewVolume;
              end else begin
                exit;
              end;
      end;
    _hi_OnEvent(_event_onNext, NrFiles);
    end;
  Close(FP);
end;


procedure THIRarInfo._work_doReadFileRar;
var Value: string;
begin
  Value := ToString(_Data);
  if FileExists(Value) then
  Read(Value);
end;

procedure THIRarInfo._var_FileName;
var N: PChar;
begin
  N := PChar(Files[NrFiles].FileName);
  OemToChar(N, N);
  dtString(_Data, N);
end;

procedure THIRarInfo._var_DateTime;
begin
  dtString(_Data, Date2StrFmt('dd-MM-yyyy',Files[NrFiles].DateTime));
end;

procedure THIRarInfo._var_Packed;
begin
  dtString(_Data, Frmt(Int2Str(Files[NrFiles].Comp_Size)));
end;

procedure THIRarInfo._var_UnPacked;
begin
  dtString(_Data, Frmt(Int2Str(Files[NrFiles].UnComp_Size)));
end;

procedure THIRarInfo._var_CompressionMethod;
begin
  dtString(_Data, Files[NrFiles].MethodStr);
end;

procedure THIRarInfo._var_CRC32;
begin
  dtString(_Data, Int2Hex(Files[NrFiles].CRC, 8));
end;

procedure THIRarInfo._var_FileTypeInfo;
begin
  dtString(_Data, GetFileTypeInfo(ExtractFileName(Files[NrFiles].FileName)));
end;

end.
