unit KOLBAPShellBrowser;

interface

uses Windows, Messages, KOL;

type
  TOnShellSelItem = procedure(Ctl: PControl; Txt: String) of object;

  PBAPShellB = ^TBAPShellB;
  TBAPShellB = object(TControl)
  private
   function GetShellErr: Boolean;
   function GetSelFile: String;
   function GetSelFolder: String;

   procedure SetShowUpDir(const Value: Boolean);
   procedure SetShowOnlyFiles(const Value: Boolean);
   procedure SetShellSort(const Value: TSortDirRules);
   procedure SetShellFilter(const Value: String);

   function GetShellPath: String;
   procedure SetShellPath(const Value: String);

   function GetNewStyle: TListViewStyle;
   procedure SetNewStyle(const Value: TListViewStyle);

   function GetOnChangeDir: TOnEvent;
   procedure SetOnChangeDir(const Value: TOnEvent);

   function GetOnShellSelItem: TOnShellSelItem;
   procedure SetOnShellSelItem(const Value: TOnShellSelItem);

   function GetNewOnLVStateChange: TOnLVStateChange;
   procedure SetNewOnLVStateChange(const Value: TOnLVStateChange);

   procedure OnNewLVStateChange(Sender: PControl; IdxFrom, IdxTo: Integer;
                                OldState, NewState: DWORD);

   procedure OnNewLVData(Sender: PControl; Idx, SubItem: Integer; var Txt: String;
                 var ImgIdx: Integer; var State: DWORD; var Store: Boolean);

  public
   property ShellErr: Boolean read GetShellErr;
   property ShowUpDir: Boolean write SetShowUpDir;
   property SelFile: String read GetSelFile;
   property SelFolder: String read GetSelFolder;
   property ShellFilter: String write SetShellFilter;
   property ShellSort: TSortDirRules write SetShellSort;
   property ShowOnlyFiles: Boolean write SetShowOnlyFiles;
   property ShellPath: String read GetShellPath write SetShellPath;
   property Style: TListViewStyle read GetNewStyle write SetNewStyle;
   property OnChangeDir: TOnEvent read GetOnChangeDir write SetOnChangeDir;
   property OnShellSelItem: TOnShellSelItem read GetOnShellSelItem write SetOnShellSelItem;
   property OnLVStateChange: TOnLVStateChange
             read GetNewOnLVStateChange write SetNewOnLVStateChange;
   property OnLVData: Boolean read FNotAvailable;

   procedure LevelUp;
   procedure RootDir;
   procedure ShellUpdate;
   procedure ShellSelect(NotSelect: Boolean);
   function Execute(Idx: Integer): Integer;
  end;

  TBAPShellBrowser = PBAPShellB;

function NewBAPShellBrowser(AParent: PControl; Style: TListViewStyle;
                            tn, ts, td: String; cn, cs, cd: Integer): PBAPShellB;

implementation

uses  {CommCtrl,} ShellApi;

{    (  ) }
type
 PShellData = ^TShellData;
 TShellData = object(TObj)
  IL16, IL32: PImageList; {  16x16  32x32 }
  DL: PDirList; {  TDirList }
  LastDirName: String; {    }
  LastDir,
  UpDir: Boolean;{  '[..]' }
  IcOF: Integer; {   ' ' }
  DrvPath: array['a'..'z'] of string; {     }

  fControl: PControl;
  SelFile: String;   {   }
  SelFolder: String; {   }
  ShellPath: String; {      }
  ShowUpDir: Boolean;{ -//-  '[..]' }
  ShellErr: Boolean; {    }
  ShowOnlyFiles: Boolean;{    }
  ShellFilter: String;   {    }
  Style: TListViewStyle; {  ListView }
  ShellSort: TSortDirRules;  {   }
  OnChangeDir: TOnEvent; {   }
  OnShellSelItem: TOnShellSelItem; {   }
  OnLVStateChange: TOnLVStateChange;
  destructor Destroy; virtual;
 end;

{-------------------------}
{ Destructor   }
{-------------------------}
destructor TShellData.Destroy;
var ch: Char;
begin
for ch := 'a' to 'z' do DrvPath[ch] := '';
SelFile := '';
SelFolder := '';
ShellPath := '';
ShellFilter := '';
LastDirName := '';
Free_And_Nil(DL);
Free_And_Nil(IL16);
Free_And_Nil(IL32);
 inherited;
end;
////////////////////////////////////////////////////////////////////////////////

{-----------------------}
{    }
{-----------------------}
function GetFileIcon(Path: String; OpenIcon: Boolean): Integer;
var
 SFI: TShFileInfo;
 Flags: Integer;
begin
Flags := SHGFI_ICON or SHGFI_SMALLICON or SHGFI_SYSICONINDEX;
if OpenIcon
 then Flags := Flags or SHGFI_OPENICON;

FillChar(SFI, sizeof(SFI), 0);
ShGetFileInfo(PChar(Path), 0, SFI, SizeOf(SFI), Flags);
Result := SFI.iIcon;
end;

{--------------------}
{    }
{--------------------}
function GetFileType(Path: String): String;
var SFI: TShFileInfo;
begin
FillChar(SFI, sizeof(SFI), 0);
ShGetFileInfo(PChar(Path), 0, SFI, SizeOf(SFI), SHGFI_TYPENAME);
Result := SFI.szTypeName;
end;
////////////////////////////////////////////////////////////////////////////////

{--------------}
{   }
{--------------}
function TBAPShellB.Execute(Idx: Integer): Integer;
var
 D: PShellData;
 st: String;
begin
D := Pointer(CustomObj);
Result := 0; { Rsult < 32 - ERROR }
st := ShellPath + D.fControl.LVItems[Idx, 0];
if FileExists(st)
 then Result := ShellExecute(Applet.Handle, nil, PChar(st), nil, nil, SW_RESTORE);
end;

{--------------------}
{    }
{--------------------}
procedure TBAPShellB.ShellSelect;
var
 D: PShellData;
 st: String;
 x, y, idx:integer;
begin
D := Pointer(CustomObj);
idx := 0;
if D.UpDir then idx := idx + 1;

{    }
x := D.fControl.LVCurItem;
y := x;

{    }
while x >= 0 do
 begin
  if (D.UpDir) and (x = 0)
   then st := ''
    else st := D.DL.Path + D.DL.Names[x - idx];

  if Assigned(D.OnShellSelItem) then D.OnShellSelItem(D.fControl, st);
  if NotSelect then D.fControl.LVItemState[x] := [];
  x := D.fControl.LVNextSelected(x);
  if x > 0 then y := x;
 end;

D.fControl.LVCurItem := y;
end;
////////////////////////////////////////////////////////////////////////////////

{-------------------------------}
{    }
{-------------------------------}
procedure GetLastDir(Ctl: PControl; var s: String);
var
 D: PShellData;
 x, y: Integer;
begin
D := Pointer(Ctl.CustomObj);
y := 1; {  LastDirName }
D.LastDirName := '';
s := ExcludeTrailingPathDelimiter(s); {  '\'}

for x := 1 to Length(s) do {   }
 if s[x] = '\' then y := x + 1;

for x := y to Length(s) do D.LastDirName := D.LastDirName + s[x];

D.LastDir := TRUE;
s := ExcludeTrailingPathDelimiter(s); {  '\'}

{   }
x := Length(s);
while s[x] <> '\' do
begin
 Delete(s, x, 1);
 x := x - 1;
end;
end;
////////////////////////////////////////////////////////////////////////////////

{------------------------}
{    }
{------------------------}
function GetNewShell(Ctl: PControl): String;
var D: PShellData;
begin
D := Pointer(Ctl.CustomObj);
D.LastDir := FALSE;
Result := D.ShellPath;
Result := IncludeTrailingPathDelimiter(Result);

if Ctl.LVItems[Ctl.LVCurItem, 0] = '..'
 then GetLastDir(Ctl, Result)
  else Result := Result + Ctl.LVItems[Ctl.LVCurItem, 0];
end;
////////////////////////////////////////////////////////////////////////////////

{----------------}
{   }
{----------------}
procedure ReadShell(Ctl: PControl; Path: String);
var
 D: PShellData;
 x, y: Integer;
 attr: DWord;
 ch: Char;
 st: String;
begin
D := Pointer(Ctl.CustomObj);

if not DirectoryExists(Path)
 then begin
  D.ShellErr := TRUE;
  if Assigned(D.OnChangeDir) then D.OnChangeDir(Ctl);
  exit;
 end;

{      }
st := LowerCase(Path);
ch := st[1];
if Length(st) = 2 then
 if D.DrvPath[ch] <> '' then Path := D.DrvPath[ch];

Path := IncludeTrailingPathDelimiter(Path);

D.ShellPath := Path;
D.DrvPath[ch] := Path;

if (Length(D.ShellPath) >= 4) and (D.ShowUpDir) and (not D.ShowOnlyFiles)
 then D.UpDir := TRUE
  else D.UpDir := FALSE;

if D.ShowOnlyFiles
 then attr := File_ATTRIBUTE_NORMAL and FILE_ATTRIBUTE_DIRECTORY
  else attr := File_ATTRIBUTE_NORMAL or FILE_ATTRIBUTE_DIRECTORY;

Ctl.BeginUpdate; { BAPShellBrowser update }
if D.DL = nil
 then D.DL := NewDirListEx(D.ShellPath, D.ShellFilter, attr)
  else D.DL.ScanDirectoryEx(Path, D.ShellFilter, attr);
D.DL.Sort(D.ShellSort);

{      }
D.IcOF := 0;
for x := 0 to D.DL.Count - 1 do D.DL.Items[x].dwReserved0 := 0;

if D.UpDir {  [..] }
 then Ctl.LVCount := D.DL.Count + 1
  else Ctl.LVCount := D.DL.Count;
Ctl.EndUpdate;

{      }
y := 0;
if D.LastDir then
for x := 0 to D.DL.Count - 1 do
 if D.LastDirName = D.DL.Names[x]
  then begin
   y := x;
   if D.UpDir then y := y + 1;
   break;
  end;

Ctl.LVCurItem := y;
Ctl.LVMakeVisible(y + Ctl.LVPerPage div 2, TRUE);
Ctl.LVMakeVisible(y, FALSE);

{  }
D.ShellErr := FALSE;
if Assigned(D.OnChangeDir) then D.OnChangeDir(Ctl);
end;
////////////////////////////////////////////////////////////////////////////////

{--------------------}
{   }
{--------------------}
procedure TBAPShellB.ShellUpdate;
var D: PShellData;
begin
D := Pointer(CustomObj);
ReadShell(D.fControl, D.ShellPath);
end;
////////////////////////////////////////////////////////////////////////////////

{-------------------------}
{     }
{-------------------------}
procedure LevelUp2(Ctl: PControl);
var
 D: PShellData;
 s: String;
begin
D := Pointer(Ctl.CustomObj);
s := D.ShellPath;
if Length(s) <= 3 then exit;
GetLastDir(Ctl, s);
ReadShell(Ctl, s);
end;

procedure TBAPShellB.LevelUp;
var D: PShellData;
begin
D := Pointer(CustomObj);
LevelUp2(D.fControl);
end;
////////////////////////////////////////////////////////////////////////////////

{------------------}
{    }
{------------------}
procedure RootDir2(Ctl: PControl);
var
 D: PShellData;
 s: String;
begin
D := Pointer(Ctl.CustomObj);
s := D.ShellPath;
if Length(s) <= 3 then exit;
Delete(s, 3, Length(s) - 3);
ReadShell(Ctl, s);
end;

procedure TBAPShellB.RootDir;
var D: PShellData;
begin
D := Pointer( CustomObj );
RootDir2(D.fControl);
end;
////////////////////////////////////////////////////////////////////////////////

{--------------------------------------}
{       }
{--------------------------------------}
function GetFileDate(Ctl: PControl; Ind, Idx: Integer): String;
var
 D: PShellData;
 ST: TSystemTime;
begin
D := Pointer(Ctl.CustomObj);

if D.UpDir and (Idx = 0)
 then begin {  ".." }
  Result := '';
  exit;
 end;

FileTimeToLocalFileTime(D.DL.Items[Ind].ftLastWriteTime, D.DL.Items[Ind].ftCreationTime);
FileTimeToSystemTime(D.DL.Items[Ind].ftCreationTime, ST);
Result :=
 SystemDate2Str(ST, LOCALE_USER_DEFAULT, dfLongDate, 'dd.MM.yy') +
 SystemTime2Str(ST, LOCALE_USER_DEFAULT, [], ' HH:mm');
end;
////////////////////////////////////////////////////////////////////////////////

{---------------------}
{  OnLVData }
{---------------------}
procedure TBAPShellB.OnNewLVData;
var
 D: PShellData;
 Ind, IC: Integer;
begin
D := Pointer(CustomObj);
if D.UpDir
 then Ind := Idx - 1
  else Ind := Idx;

case SubItem of
0: //   //
begin
 if D.UpDir and (Idx = 0)
  then begin {  ".." }
   Txt := '..';
   if D.IcOF = 0 then D.IcOF := GetFileIcon(D.ShellPath, TRUE);
   ImgIdx := D.IcOF; {     }
   exit
  end;

Txt := D.DL.Names[Ind]; {    }

{     }
if D.DL.Items[ind].dwReserved0 = 0
 then begin
  IC := GetFileIcon(ShellPath + D.DL.Names[ind], FALSE);
  if IC = 0 then IC := FileIconSysIdxOffline(ShellPath + D.DL.Names[ind]);
  D.DL.Items[ind].dwReserved0 := IC;
 end;

ImgIdx := D.DL.Items[ind].dwReserved0;
end;

1: //   //
begin
if (Sender.LVColWidth[1] <= 0) then exit;
if D.UpDir and (Idx = 0)
 then begin {  ".." }
  Txt := '';
  exit;
 end;

if DirectoryExists(D.ShellPath + D.DL.Names[ind])
 then Txt := GetFileType(D.ShellPath + D.DL.Names[ind])
  else Txt := Num2Bytes(Int64_2Double(MakeInt64(D.DL.Items[Ind].nFileSizeLow, D.DL.Items[Ind].nFileSizeHigh)));
 end;

2: //   //
begin
if (Sender.LVColWidth[2] <= 0) then exit;
Txt := GetFileDate(Sender, Ind, Idx);
end;
end; { case }

Store := FALSE;
end;
////////////////////////////////////////////////////////////////////////////////

{---------------------------------}
{      }
{---------------------------------}
procedure GetNameFF(Ctl: PControl; st: String);
var D: PShellData;
begin
D := Pointer(Ctl.CustomObj);
D.SelFolder := '';
D.SelFile := '';

if FileExists(D.ShellPath  + st) {    }
 then D.SelFile := D.ShellPath  + st;

if DirectoryExists(D.ShellPath + st) {    }
 then D.SelFolder := D.ShellPath + st + '\';

if st = '..' then D.SelFolder := D.ShellPath;
end;
////////////////////////////////////////////////////////////////////////////////

{----------------------------}
{  OnLVStateChange }
{----------------------------}
procedure TBAPShellB.OnNewLVStateChange;
var D: PShellData;
begin
D := Pointer(CustomObj);

if D.fControl.LVItems[D.fControl.LVCurItem, 0] <> ''
 then GetNameFF(D.fControl, D.fControl.LVItems[IdxTo, 0]);

if Assigned(D.OnLVStateChange)
 then D.OnLVStateChange(@Self, IdxFrom, IdxTo, OldState, NewState);
end;
////////////////////////////////////////////////////////////////////////////////

{-----------------------------------------}
{   myOnLVStateChange }
{-----------------------------------------}
procedure TBAPShellB.SetNewOnLVStateChange;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.OnLVStateChange := Value;
end;

function TBAPShellB.GetNewOnLVStateChange;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.OnLVStateChange;
end;
////////////////////////////////////////////////////////////////////////////////

{--------------------}
{   }
{--------------------}
function WndProcShell(Ctl: PControl; var Msg: TMsg; var Rslt: Integer): Boolean;
//var LV: PNMLISTVIEW;
begin
Result := FALSE;
case  Msg.message of

WM_KEYDOWN:
 case Msg.wParam of
  220: if GetKeyState(VK_CONTROL) < 0
        then RootDir2(Ctl); {  "\" }

  VK_RETURN: ReadShell(Ctl, GetNewShell(Ctl)); {  Enter }
  VK_BACK: LevelUp2(Ctl); {  Backspace }

  {    }
  VK_LEFT:
   if (Ctl.LVStyle = lvsDetail) or (Ctl.LVStyle = lvsDetailNoHeader)
    then Ctl.Perform(WM_KEYDOWN, VK_PRIOR, 0);

  VK_RIGHT:
   if (Ctl.LVStyle = lvsDetail) or (Ctl.LVStyle = lvsDetailNoHeader)
    then Ctl.Perform(WM_KEYDOWN, VK_NEXT, 0);
  end; { Msg.wParam }

{WM_NOTIFY:
begin
 LV := Pointer(Msg.lParam);

 if LV.hdr.code = NM_DBLCLK  //      
  then ReadShell(Ctl, GetNewShell(Ctl));
end; // WM_NOTIFY }
end; //case }
end;
////////////////////////////////////////////////////////////////////////////////

{-----------------------------}
{   KOL  }
{-----------------------------}
function NewBAPShellBrowser;
var D: PShellData;
begin
New(D, Create);

D.IL16 := NewImageList(nil);
D.IL16.LoadSystemIcons(TRUE);
D.IL32 := NewImageList(nil);
D.IL32.LoadSystemIcons(FALSE);

{   }
Result := PBAPShellB(NewListView(AParent, Style,
                     [lvoOwnerData, lvoRowSelect, lvoMultiSelect], D.IL16, D.IL32, nil));

Result.CustomObj := D;
D.fControl := Result;

{    }
D.ShowUpDir := TRUE;
D.ShowOnlyFiles := FALSE;
D.ShellSort := sdrByExt;
D.ShellFilter := '*.*';

if cn <= 0 then cn := 200;
Result.LVColInsert(0, tn, taLeft, cn);
Result.LVColInsert(1, ts, taRight, cs);
Result.LVColInsert(2, td, taLeft, cd);

if cs <= 0 then Result.LVColWidth[1] := 0;
if cd <= 0 then Result.LVColWidth[2] := 0;

{   }
Result.SetOnLVStateChange(Result.OnNewLVStateChange);
Result.SetOnLVData(Result.OnNewLVData);
Result.AttachProc(WndProcShell);
end;
////////////////////////////////////////////////////////////////////////////////

{----------------------------}
{    }
{----------------------------}
procedure TBAPShellB.SetShellPath;
var
 D: PShellData;
 s: String;
begin
D := Pointer(CustomObj);
s := Value;
if s = '*' then s := GetStartDir;
ReadShell(D.fControl, s);
end;

function TBAPShellB.GetShellPath;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.ShellPath;
end;
////////////////////////////////////////////////////////////////////////////////

{------------------}
{  Style }
{------------------}
procedure TBAPShellB.SetNewStyle;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.Style := Value;
D.fControl.LVStyle := Value;
if Value = lvsList
 then ReadShell(D.fControl, D.ShellPath);
end;

function TBAPShellB.GetNewStyle;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.Style;
end;
////////////////////////////////////////////////////////////////////////////////

{------------------------}
{  OnChangeDir }
{------------------------}
procedure TBAPShellB.SetOnChangeDir;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.OnChangeDir := Value;
end;

function TBAPShellB.GetOnChangeDir;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.OnChangeDir;
end;

{------------------------}
{  OnNewSelect }
{------------------------}
procedure TBAPShellB.SetOnShellSelItem;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.OnShellSelItem := Value;
end;

function TBAPShellB.GetOnShellSelItem;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.OnShellSelItem;
end;
////////////////////////////////////////////////////////////////////////////////

{-----------------}
{  ShowUpDir }
{-----------------}
procedure TBAPShellB.SetShowUpDir;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.ShowUpDir := Value;
ReadShell(D.fControl, D.ShellPath);
end;

{-----------------------}
{    }
{-----------------------}
procedure TBAPShellB.SetShowOnlyFiles;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.ShowOnlyFiles := Value;
ReadShell(D.fControl, D.ShellPath);
end;

{-------------------}
{   }
{-------------------}
procedure TBAPShellB.SetShellSort;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.ShellSort := Value;
ReadShell(D.fControl, D.ShellPath);
end;

{-------------------}
{   }
{-------------------}
procedure TBAPShellB.SetShellFilter;
var D: PShellData;
begin
D := Pointer(CustomObj);
D.ShellFilter := Value;
ReadShell(D.fControl, D.ShellPath);
end;

{-------------------------}
{    }
{-------------------------}
function TBapShellB.GetSelFile;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.SelFile;
end;

{----------------------------}
{    }
{----------------------------}
function TBapShellB.GetSelFolder;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.SelFolder;
end;

{------------------------}
{    }
{------------------------}
function TBapShellB.GetShellErr;
var D: PShellData;
begin
D := Pointer(CustomObj);
Result := D.ShellErr;
end;
////////////////////////////////////////////////////////////////////////////////

end.
