unit hiStackEx;

interface

uses Kol,Share,Debug;

type
  THIStackEx = class(TDebug)
   private
    FData:TData;
    FDefault:TData;
    FList:PList;
    MaxSize:integer;

    procedure SetDefault(Data:TData);
    procedure SetMaxSize(const Value:integer);
   public
    _prop_StackType:byte;
    _prop_IgnorEmpty:boolean;
    _prop_Overflow:byte;
    
    _data_Data:THI_Event;
    _event_onPush:THI_Event;
    _event_onPop:THI_Event;
    _event_onEmpty:THI_Event;
    _event_onOverflow:THI_Event;

    constructor Create;
    destructor Destroy; override;
    procedure _work_doPush0(var _Data:TData; Index:word);
    procedure _work_doPush1(var _Data:TData; Index:word);
    procedure _work_doPushBack(var _Data:TData; Index:word);
    procedure _work_doPushFront(var _Data:TData; Index:word);
    procedure _work_doPop(var _Data:TData; Index:word);
    procedure _work_doClear(var _Data:TData; Index:word);
    procedure _var_Value(var _Data:TData; Index:word);
    procedure _var_Count(var _Data:TData; Index:word);
    property _prop_Default:TData write SetDefault;
    property _prop_MaxSize:integer write SetMaxSize;
  end;

implementation

constructor THIStackEx.Create;
begin
  inherited;
  FList := newlist;
end;

destructor THIStackEx.Destroy;
begin
  FList.Free;
  inherited;
end;

procedure THIStackEx.SetDefault;
begin
   FDefault := Data;
   FData := FDefault;
end;

procedure THIStackEx.SetMaxSize;
begin
  MaxSize := Value;
end;

procedure THIStackEx._work_doPush0;  //FIFO
begin
  _work_doPushFront(_Data, Index);
end;

procedure THIStackEx._work_doPushFront; //FIFO
var dt:PData;
begin
  if (MaxSize = 0) or (MaxSize > FList.Count) then
    begin
      FData := ReadData(_Data,_data_Data);
      new(dt);
      dtData(dt^,FData);
      FList.Insert(0,dt);
      _hi_CreateEvent_(_Data,@_event_onPush);
    end
  else
    begin
      if (_prop_Overflow = 1) then
        begin
          FList.Delete(FList.Count-1);
          FData := ReadData(_Data,_data_Data);
          new(dt);
          dtData(dt^,FData);
          FList.Insert(0,dt);
        end;
      _hi_CreateEvent_(_Data,@_event_onOverflow);
    end;
end;

procedure THIStackEx._work_doPush1; //LIFO
begin
  _work_doPushBack(_Data, Index);
end;

procedure THIStackEx._work_doPushBack; //LIFO
var dt:PData;
begin
  if (MaxSize = 0) or (MaxSize > FList.Count) then
    begin
      FData := ReadData(_Data,_data_Data);
      new(dt);
      dtData(dt^,FData);
      FList.Add(dt);
      _hi_CreateEvent_(_Data,@_event_onPush);
    end
  else
    begin
      if (_prop_Overflow = 1) then
        begin
          FList.Delete(0);
  	      FData := ReadData(_Data,_data_Data);
  	      new(dt);
	      dtData(dt^,FData);
	      FList.Add(dt);
	    end;
      _hi_CreateEvent_(_Data,@_event_onOverflow);
    end;
end;

procedure THIStackEx._work_doPop;
var dt:PData;
begin
  if FList.Count > 0 then begin
    dt := PData(FList.Items[FList.Count-1]);
    FData := dt^;
    dispose(dt);
    FList.Delete(FList.Count-1);
  end else if _prop_IgnorEmpty then begin
    _hi_CreateEvent(_Data,@_event_onEmpty);
    Exit;
  end else FData := FDefault;
  _hi_CreateEvent(_Data,@_event_onPop,FData);
end;

procedure THIStackEx._work_doClear;
var i:integer;
begin
  for i := 0 to FList.Count-1 do
    dispose(pdata(FList.Items[i]));
  FList.Clear;
end;

procedure THIStackEx._var_Value;
begin
  _Data := FData;
end;

procedure THIStackEx._var_Count;
begin
  dtInteger(_Data, FList.Count);
end;
end.
